#cs(module parse-error mzscheme
(require "myenv.ss")
;; This code provides error and warnings handling for SSAX (S)XML parser
;  for PLT Schemes 2.xx 
;  $Id: parse-error.scm,v 2.1 2003/12/20 02:50:07 kl Exp kl $

;==============================================================================
; Error handler

; According to the SSAX convention this function
; accepts the port as its first argument which is used for
; location of the error in input file.
; Other parameters are considered as error messages,
;  they are printed to stderr as is.
; Error message provide the information about input port 
;  (absolute file path name for file ports) and position
; If 
;   (port-count-lines! <port>) 
;  was called following the opening of an input port 
; then  additionall information about line and column 
;  of the error location is also included

; Just a wrapper: regular port -> lines-counting port 
(define (port-count-lines p)
  (port-count-lines! p)
  p)

(define parser-error-default
  (lambda  (port . messages)
    (if (port? port)
      (call-with-values 
	(lambda() (port-next-location port))
	(lambda (line col pos)
	  (cerr nl "Error in "
		(object-name port))
	  (if line
	    (cerr " at line " line " column " col)
	    (cerr " at")) 
	  (cerr " position " pos nl messages nl)))
      (cerr nl "Error in error handler: " port
	    "is not a port!" nl messages nl))
    (exit -1)))

; Default warnings handler: unlike errors handler it doesn't exit
(define ssax:warn-default
  (lambda  (port . messages)
    (if
      (port? port)
      (call-with-values 
	(lambda() (port-next-location port))
	(lambda (line col pos)
	  (cerr nl "Warning in "
		(object-name port))
	  (if line
	    (cerr " at line " line " column " col)
	    (cerr " at")) 
	  (cerr " position " pos nl messages nl)))
      (cerr nl "Error in warnings handler: " port
	    "is not a port!" nl messages nl))
      #f))

;============================================================================
; Handlers parameterisation and modification 
 

; `parser-error' is a procedure that dispatches to the value of
; the `parser-error-param' parameter.  The dispatching procedure
; lets us avoid modifying the call points to `parser-error' in the
; SSAX code.  An added benefit of using parameters is that
; two different threads can use the same SSAX code simultaneously
; with different error handlers.  For convenience, the
; `parser-error-param' parameter defaults to `parser-error-default'
; procedure.  All modules that access `parser-error' must import
; this module, and must not redefine `parser-error'.
(define parser-error-param (make-parameter parser-error-default))

(define parser-error (lambda args (apply (parser-error-param) args)))

; This function changes default handler for parser errors  
(define (set-parser-error! handler)
    (set! parser-error-param (make-parameter handler)))

; Warnings handler parameter
(define ssax:warn-param (make-parameter ssax:warn-default))

(define ssax:warn (lambda args (apply (ssax:warn-param) args)))

; This function changes default handler for parser warnongs
(define (ssax:set-warn! handler)
    (set! ssax:warn-param (make-parameter handler)))

(provide (all-defined)))
